@@ -49,6 +49,12 @@ HTMLObjectClass html_object_class;
 
 /* HTMLObject virtual methods.  */
 
+static void
+free_data (GQuark id, gpointer data, gpointer user_data)
+{
+	g_free (data);
+}
+
 static void
 destroy (HTMLObject *self)
 {
@@ -63,6 +69,7 @@ destroy (HTMLObject *self)
 		self->next = NULL;
 		self->prev = NULL;
 #endif
+		g_datalist_foreach (&self->object_data, free_data, NULL);
 		g_datalist_clear (&self->object_data);
 		g_free (self);
 	}
@@ -118,12 +125,12 @@ op_cut (HTMLObject *self, HTMLEngine *e, GList *from, GList *to, GList *left, GL
 			if (e->cursor->object == self)
 				e->cursor->object = empty;
 			html_clue_append_after (HTML_CLUE (self->parent), empty, self);
-			html_object_change_set (empty, HTML_CHANGE_ALL);
+			html_object_change_set (empty, HTML_CHANGE_ALL_CALC);
 			html_object_check_cut_lists (self, empty, left, right);
 		} else
 			html_object_move_cursor_before_remove (self, e);
 
-		html_object_change_set   (self,  HTML_CHANGE_ALL);
+		html_object_change_set   (self,  HTML_CHANGE_ALL_CALC);
 		html_object_remove_child (self->parent, self);
 		*len += html_object_get_recursive_length (self);
 
@@ -262,19 +269,19 @@ set_max_width (HTMLObject *o, HTMLPainter *painter, gint max_width)
 }
 
 static gint
-get_left_margin (HTMLObject *self, gint y)
+get_left_margin (HTMLObject *self, HTMLPainter *painter, gint y)
 {
 	return 0;
 }
 
 static gint
-get_right_margin (HTMLObject *self, gint y)
+get_right_margin (HTMLObject *self, HTMLPainter *painter, gint y)
 {
 	return MAX (self->max_width, self->width);
 }
 
 static void
-set_painter (HTMLObject *o, HTMLPainter *painter, gint max_width)
+set_painter (HTMLObject *o, HTMLPainter *painter)
 {
 }
 
@@ -298,6 +305,12 @@ get_target (HTMLObject *o)
 	return NULL;
 }
 
+static const gchar *
+get_src (HTMLObject *o)
+{
+	return NULL;
+}
+
 static HTMLAnchor *
 find_anchor (HTMLObject *o,
 	     const gchar *name,
@@ -470,6 +483,12 @@ get_length (HTMLObject *self)
 	return 1;
 }
 
+static guint
+get_line_length (HTMLObject *self, HTMLPainter *p, gint line_offset)
+{
+	return html_object_get_length (self);
+}
+
 static guint
 get_recursive_length (HTMLObject *self)
 {
@@ -619,6 +638,7 @@ html_object_class_init (HTMLObjectClass *klass,
 	klass->calc_preferred_width = calc_preferred_width;
 	klass->get_url = get_url;
 	klass->get_target = get_target;
+	klass->get_src = get_src;
 	klass->find_anchor = find_anchor;
 	klass->set_link = NULL;
 	klass->set_bg_color = set_bg_color;
@@ -639,6 +659,7 @@ html_object_class_init (HTMLObjectClass *klass,
 	klass->search = search;
 	klass->search_next = search;
 	klass->get_length = get_length;
+	klass->get_line_length = get_line_length;
 	klass->get_recursive_length = get_recursive_length;
 	klass->next = next;
 	klass->prev = prev;
@@ -877,42 +898,27 @@ html_object_set_max_width (HTMLObject *o, HTMLPainter *painter, gint max_width)
 }
 
 gint
-html_object_get_left_margin (HTMLObject *self, gint y)
+html_object_get_left_margin (HTMLObject *self, HTMLPainter *painter, gint y)
 {
-	return (* HO_CLASS (self)->get_left_margin) (self, y);
+	return (* HO_CLASS (self)->get_left_margin) (self, painter, y);
 }
 
 gint
-html_object_get_right_margin (HTMLObject *self, gint y)
+html_object_get_right_margin (HTMLObject *self, HTMLPainter *painter, gint y)
 {
-	return (* HO_CLASS (self)->get_right_margin) (self, y);
+	return (* HO_CLASS (self)->get_right_margin) (self, painter, y);
 }
 
-struct _hlpSetPainter {
-	HTMLPainter *painter;
-	gint max_width;
-};
-typedef struct _hlpSetPainter hlpSetPainter;
-
 static void
 set_painter_forall (HTMLObject *o, HTMLEngine *e, gpointer data)
 {
-	hlpSetPainter *sp;
-
-	sp = (hlpSetPainter *) data;
-	(* HO_CLASS (o)->set_painter) (o, sp->painter, sp->max_width);
+	(* HO_CLASS (o)->set_painter) (o, HTML_PAINTER (data));
 }
 
 void
-html_object_set_painter (HTMLObject *o, HTMLPainter *painter, gint max_width)
+html_object_set_painter (HTMLObject *o, HTMLPainter *painter)
 {
-	hlpSetPainter *sp;
-
-	sp = g_new (hlpSetPainter, 1);
-	sp->painter = painter;
-	sp->max_width = max_width;
-	html_object_forall (o, NULL, set_painter_forall, sp);
-	g_free (sp);
+	html_object_forall (o, NULL, set_painter_forall, painter);
 }
 
 void
@@ -943,6 +949,11 @@ html_object_calc_preferred_width (HTMLObject *o,
 	return o->pref_width;
 }
 
+gint
+html_object_get_uris (HTMLObject *o, char **link, char **target, char **src) {
+	return TRUE;
+}
+
 const gchar *
 html_object_get_url (HTMLObject *o)
 {
@@ -955,6 +966,12 @@ html_object_get_target (HTMLObject *o)
 	return (* HO_CLASS (o)->get_target) (o);
 }
 
+const gchar *
+html_object_get_src (HTMLObject *o)
+{
+	return (* HO_CLASS (o)->get_src) (o);
+}
+
 HTMLAnchor *
 html_object_find_anchor (HTMLObject *o,
 			 const gchar *name,
@@ -1212,7 +1229,7 @@ HTMLObject *
 html_object_remove_link (HTMLObject *self,
 			 HTMLColor *color)
 {
-	return (HO_CLASS (self)->remove_link) ? (* HO_CLASS (self)->remove_link) (self, color) : NULL;
+	return (HO_CLASS (self)->set_link) ? (* HO_CLASS (self)->set_link) (self, color, NULL, NULL) : NULL;
 }
 
 guint
@@ -1221,6 +1238,12 @@ html_object_get_length (HTMLObject *self)
 	return (* HO_CLASS (self)->get_length) (self);
 }
 
+guint
+html_object_get_line_length (HTMLObject *self, HTMLPainter *p, gint line_offset)
+{
+	return (* HO_CLASS (self)->get_line_length) (self, p, line_offset);
+}
+
 guint
 html_object_get_recursive_length (HTMLObject *self)
 {
@@ -1490,9 +1513,9 @@ html_object_get_index (HTMLObject *self, guint offset)
 }
 
 void
-html_object_set_data (HTMLObject *object, const gchar *key, gpointer data)
+html_object_set_data (HTMLObject *object, const gchar *key, const gchar *value)
 {
-	g_datalist_set_data (&object->object_data, key, data);
+	g_datalist_set_data (&object->object_data, key, g_strdup (value));
 }
 
 gpointer
@@ -1506,7 +1529,7 @@ copy_data (GQuark key_id, gpointer data, gpointer user_data)
 {
 	HTMLObject *o = HTML_OBJECT (user_data);
 
-	g_datalist_id_set_data (&o->object_data, key_id, data);
+	g_datalist_id_set_data (&o->object_data, key_id, g_strdup ((gchar *)data));
 }
 
 void
@@ -1518,7 +1541,7 @@ html_object_copy_data_from_object (HTMLObject *dst, HTMLObject *src)
 GList *
 html_object_get_bound_list (HTMLObject *self, GList *list)
 {
-	return list
+	return list && list->next
 		? (HTML_OBJECT (list->data) == self ? list->next : NULL)
 		: NULL;
 }
@@ -1658,3 +1681,68 @@ html_object_nth_parent (HTMLObject *self, gint n)
 
 	return self;
 }
+
+gint
+html_object_get_parent_level (HTMLObject *self)
+{
+	gint level = 0;
+
+	while (self) {
+		level ++;
+		self = self->parent;
+	}
+
+	return level;
+}
+
+GList *
+html_object_heads_list (HTMLObject *o)
+{
+	GList *list = NULL;
+
+	g_return_val_if_fail (o, NULL);
+
+	while (o) {
+		list = g_list_append (list, o);
+		o = html_object_head (o);
+	}
+
+	return list;
+}
+
+GList *
+html_object_tails_list (HTMLObject *o)
+{
+	GList *list = NULL;
+
+	g_return_val_if_fail (o, NULL);
+
+	while (o) {
+		list = g_list_append (list, o);
+		o = html_object_tail_not_slave (o);
+	}
+
+	return list;
+}
+
+static void
+merge_down (HTMLEngine *e, GList *left, GList *right)
+{
+	HTMLObject *lo;
+	HTMLObject *ro;
+
+	while (left && right) {
+		lo    = HTML_OBJECT (left->data);
+		ro    = HTML_OBJECT (right->data);
+		left  = left->next;
+		right = right->next;
+		if (!html_object_merge (lo, ro, e, left, right))
+			break;
+	}
+}
+
+void
+html_object_merge_down (HTMLObject *o, HTMLObject *w, HTMLEngine *e)
+{
+	merge_down (e, html_object_tails_list (o), html_object_heads_list (w));
+}
